/*

gbmraw.cpp - Camera RAW

Credit for writing this module must go to Heiko Nitzsche.
This file is just as public domain as the rest of GBM.

Supported formats and options:
------------------------------
Camera RAW : All dcraw supported formats
(DNG 3FR ARW CRW CR2 DCR DCS KDC ERF MEF MRW MDC NEF ORF RAF RAW RW2 PEF X3F SRF SR2)

Standard formats (backward compatible):
  Reads files and presents them as 24 bpp.

Extended formats (not backward compatible, import option ext_bpp required):
  Reads files and presents them as 48 bpp.

Input:
------

Can specify that non-standard GBM color depth is exported (48 bpp)
  Input option: ext_bpp (default: bpp is downsampled to 24 bpp)

Can specify that a half-sized image is created
  Input option: halfsize (default is full size)

Can specify the interpolation quality
  Input option: quality=# (0 to 3, default 0)

Can specify that the average of the whole image is used for white balance
  Input option: avgwb  (default is camera white balance)

Can specify a custom white balance
  Input option: wb=r_g_b_g

Can specify that chromatic aberration is corrected
  Input option: aber=r_b

Can specify that the image is automatically brightened
  Input option: autobright

Can specify the image brightness
  Input option: bright=# (default: 1.0)

Can specify the threshold for wavelet denoising
  Input option: denoise=#

Can specify how often a 3x3 median filter is applied to R-G and B-G
  Input option: medianfilter=# (0..N, default 0)


History:
--------
(Heiko Nitzsche)

20-Sep-2010: Initial version based on libraw 0.10.0
*/

#ifdef ENABLE_RAW

#include <stdio.h>
#include <ctype.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include "gbm.h"
#include "gbmraw.h"
#include "gbmhelp.h"
#include "gbmdesc.h"
#include "gbmmem.h"
#include "gbmmap.h"
#include "gbmmthrd.h"

#include "libraw.h"

/* ---------------------------------------- */

#define GBM_ERR_RAW_NO_THUMB         ((GBM_ERR) 8200)
#define GBM_ERR_RAW_BAD_THUMB        ((GBM_ERR) 8201)
#define GBM_ERR_RAW_BAD_QUALITY      ((GBM_ERR) 8202)
#define GBM_ERR_RAW_BAD_ABERRATION   ((GBM_ERR) 8203)
#define GBM_ERR_RAW_BAD_BRIGHT       ((GBM_ERR) 8204)
#define GBM_ERR_RAW_BAD_BRIGHTCOMB   ((GBM_ERR) 8205)
#define GBM_ERR_RAW_BAD_WB           ((GBM_ERR) 8206)
#define GBM_ERR_RAW_BAD_WBCOMB       ((GBM_ERR) 8207)
#define GBM_ERR_RAW_BAD_DENOISE      ((GBM_ERR) 8208)
#define GBM_ERR_RAW_BAD_MEDIANFILTER ((GBM_ERR) 8209)

/* ---------------------------------------- */

static GBMFT raw_gbmft =
{
   GBM_FMT_DESC_SHORT_RAW,
   GBM_FMT_DESC_LONG_RAW,
   GBM_FMT_DESC_EXT_RAW,
   GBM_FT_R24 | GBM_FT_R48
};

typedef struct
{
   /* This entry will store the options provided during first header read.
    * It will keep the options for the case the header has to be reread.
    */
   char read_options[PRIV_SIZE - sizeof(int)
                               - 8 /* space for structure element padding */ ];
} RAW_PRIV_READ;

/* ---------------------------------------- */
/* ---------------------------------------- */

static GBM_ERR MakeGbmErrFromRawErr(const int rawErr, const GBM_ERR defaultGbmErr)
{
    GBM_ERR gbmRc = GBM_ERR_OK;
    switch(rawErr)
    {
        case LIBRAW_SUCCESS:               gbmRc = GBM_ERR_OK;            break;
        case LIBRAW_FILE_UNSUPPORTED:      gbmRc = GBM_ERR_NOT_SUPP;      break;
        case LIBRAW_NO_THUMBNAIL:          gbmRc = GBM_ERR_RAW_NO_THUMB;  break;
        case LIBRAW_UNSUPPORTED_THUMBNAIL: gbmRc = GBM_ERR_RAW_BAD_THUMB; break;
        case LIBRAW_UNSUFFICIENT_MEMORY:   gbmRc = GBM_ERR_MEM;           break;
        case LIBRAW_DATA_ERROR:            gbmRc = GBM_ERR_NOT_SUPP;      break;
        case LIBRAW_IO_ERROR:              gbmRc = GBM_ERR_READ;          break;

        default:
            gbmRc = defaultGbmErr;
            break;
    }
    return gbmRc;
}

/* ---------------------------------------- */
/* ---------------------------------------- */

GBM_ERR raw_qft(GBMFT *gbmft)
{
   *gbmft = raw_gbmft;
   return GBM_ERR_OK;
}

/* ---------------------------------------- */
/* ---------------------------------------- */

static GBM_ERR internal_raw_read(int fd, GBM * gbm, gbm_u8 * data, int *pimgcnt, gbm_boolean imgCntOnly)
{
   RAW_PRIV_READ *raw_priv = (RAW_PRIV_READ *) gbm->priv;
   const char * s      = NULL;
   gbm_u8     * buffer = NULL;
   gbm_boolean  use_native_bpp, isUnpackedHalfSize = GBM_FALSE;
   long         fsize  = 0;
   int          rawErr = 0;

   /* decode raw file via buffer */
   libraw_data_t * libraw_data = libraw_init(LIBRAW_OPTIONS_NONE);
   if (NULL == libraw_data)
   {
       return GBM_ERR_MEM;
   }

   /* check if extended color depths are requested */
   use_native_bpp = (gbm_find_word(raw_priv->read_options, "ext_bpp") != NULL)
                    ? GBM_TRUE : GBM_FALSE;

   /* set the default processing parameters */
   libraw_data->params.output_bps     = use_native_bpp ? 16 : 8;
   libraw_data->params.document_mode  = 0;
   libraw_data->params.output_color   = 1; /* sRGB */
   libraw_data->params.shot_select    = 0;
   libraw_data->params.half_size      = 0;

   libraw_data->params.gamm[0]        = 1.0/2.4;
   libraw_data->params.gamm[1]        = 12.92;
   libraw_data->params.use_auto_wb    = 0;
   libraw_data->params.use_camera_wb  = 1;
   libraw_data->params.no_auto_bright = 1;
   libraw_data->params.user_qual      = 0; /* 0 - linear, 1- VNG, 2 - PPG, 3 - AHD */

   if (gbm_find_word(raw_priv->read_options, "halfsize") != NULL)
   {
       libraw_data->params.half_size = 1;
       libraw_data->params.four_color_rgb = 1;
   }

   /* white balance settings */
   if (gbm_find_word(raw_priv->read_options, "avgwb") != NULL)
   {
       if (gbm_find_word(raw_priv->read_options, "wb=") != NULL)
       {
           return GBM_ERR_RAW_BAD_WBCOMB;
       }
       libraw_data->params.use_auto_wb   = 1;
       libraw_data->params.use_camera_wb = 0;
   }
   if ((s = gbm_find_word_prefix(raw_priv->read_options, "wb=")) != NULL)
   {
      float r = 1.0, g1 = 1.0, b = 1.0, g2 = 1.0;
      if (gbm_find_word(raw_priv->read_options, "avgwb") != NULL)
      {
          return GBM_ERR_RAW_BAD_WBCOMB;
      }
      if (sscanf(s + 3, "%f_%f_%f_%f", &r, &g1, &b, &g2) != 4)
      {
         return GBM_ERR_RAW_BAD_WB;
      }
      libraw_data->params.user_mul[0] = r;
      libraw_data->params.user_mul[1] = g1;
      libraw_data->params.user_mul[2] = b;
      libraw_data->params.user_mul[3] = g2;
   }

   /* quality related settings */
   if ((s = gbm_find_word_prefix(raw_priv->read_options, "quality=")) != NULL)
   {
       if ((sscanf(s + 8, "%d", &(libraw_data->params.user_qual)) != 1)
           || (libraw_data->params.user_qual < 0)
           || (libraw_data->params.user_qual > 3))
       {
           return GBM_ERR_RAW_BAD_QUALITY;
       }
   }
   if ((s = gbm_find_word_prefix(raw_priv->read_options, "aber=")) != NULL)
   {
      double redFactor   = 0;
      double greenFactor = 0;
      if (sscanf(s + 5, "%lf_%lf", &redFactor, &greenFactor) != 2)
      {
         return GBM_ERR_RAW_BAD_ABERRATION;
      }
      libraw_data->params.aber[0] = 1.0 / redFactor;
      libraw_data->params.aber[2] = 1.0 / greenFactor;
      isUnpackedHalfSize = GBM_TRUE;
   }
   if ((s = gbm_find_word_prefix(raw_priv->read_options, "denoise=")) != NULL)
   {
       if (sscanf(s + 8, "%f", &(libraw_data->params.threshold)) != 1)
       {
           return GBM_ERR_RAW_BAD_DENOISE;
       }
       isUnpackedHalfSize = GBM_TRUE;
   }

   if ((s = gbm_find_word_prefix(raw_priv->read_options, "medianfilter=")) != NULL)
   {
       if ((sscanf(s + 13, "%d", &(libraw_data->params.med_passes)) != 1)
           || (libraw_data->params.med_passes < 0))
       {
           return GBM_ERR_RAW_BAD_MEDIANFILTER;
       }
   }

   /* brightness settings */
   if (gbm_find_word(raw_priv->read_options, "autobright") != NULL)
   {
       if (gbm_find_word(raw_priv->read_options, "bright=") != NULL)
       {
           return GBM_ERR_RAW_BAD_BRIGHTCOMB;
       }
       libraw_data->params.no_auto_bright = 0;
   }
   if ((s = gbm_find_word_prefix(raw_priv->read_options, "bright=")) != NULL)
   {
       if (gbm_find_word(raw_priv->read_options, "autobright") != NULL)
       {
           return GBM_ERR_RAW_BAD_BRIGHTCOMB;
       }
       if (sscanf(s + 7, "%f", &(libraw_data->params.bright)) != 1)
       {
           return GBM_ERR_RAW_BAD_BRIGHT;
       }
   }

   /* read the whole file and then process from buffer */
   fsize = gbm_file_lseek(fd, 0, GBM_SEEK_END);

   /* start at the beginning of the file */
   if (gbm_file_lseek(fd, 0, GBM_SEEK_SET) < 0)
   {
       libraw_close(libraw_data);
       return GBM_ERR_READ;
   }

   buffer = (gbm_u8 *)gbmmem_malloc(fsize);
   if (buffer == NULL)
   {
       libraw_close(libraw_data);
       return GBM_ERR_MEM;
   }
   if (fsize != gbm_file_read(fd, buffer, fsize))
   {
       libraw_close(libraw_data);
       gbmmem_free(buffer);
       return GBM_ERR_READ;
   }

   rawErr = libraw_open_buffer(libraw_data, buffer, fsize);
   if (rawErr != LIBRAW_SUCCESS)
   {
       libraw_close(libraw_data);
       gbmmem_free(buffer);
       return MakeGbmErrFromRawErr(rawErr, GBM_ERR_READ);
   }

   /* only RGB supported */
   if (libraw_data->idata.colors != 3)
   {
       libraw_close(libraw_data);
       gbmmem_free(buffer);
       return GBM_ERR_NOT_SUPP;
   }

   *pimgcnt = libraw_data->idata.raw_count;
   if (*pimgcnt < 1)
   {
       /* image format not recognized */
       libraw_close(libraw_data);
       gbmmem_free(buffer);
       return GBM_ERR_NOT_SUPP;
   }
   if (imgCntOnly)
   {
       libraw_close(libraw_data);
       gbmmem_free(buffer);
       return GBM_ERR_OK;
   }

   /* Do info analysis only if no data are requested. */
   if (data == NULL)
   {
       rawErr = libraw_adjust_sizes_info_only(libraw_data);
       if (rawErr != LIBRAW_SUCCESS)
       {
           libraw_close(libraw_data);
           gbmmem_free(buffer);
           return MakeGbmErrFromRawErr(rawErr, GBM_ERR_READ);
       }

       gbm->w   = libraw_data->sizes.iwidth;
       gbm->h   = libraw_data->sizes.iheight;
       gbm->bpp = libraw_data->idata.colors * libraw_data->params.output_bps;

       if (isUnpackedHalfSize)
       {
           gbm->w = 2 * gbm->w + 1;
           gbm->h = 2 * gbm->h + 1;
       }

       libraw_close(libraw_data);
       gbmmem_free(buffer);
       return GBM_ERR_OK;
   }

   /* unfortunately we have to unpack and decode the image to get correct sizes */
   rawErr = libraw_unpack(libraw_data);
   if (rawErr != LIBRAW_SUCCESS)
   {
       libraw_close(libraw_data);
       gbmmem_free(buffer);
       return MakeGbmErrFromRawErr(rawErr, GBM_ERR_NOT_SUPP);
   }
   gbmmem_free(buffer);

   /* decode */
   {
       libraw_processed_image_t *image = NULL;

       rawErr = libraw_dcraw_process(libraw_data);
       if (rawErr != LIBRAW_SUCCESS)
       {
           libraw_close(libraw_data);
           return MakeGbmErrFromRawErr(rawErr, GBM_ERR_NOT_SUPP);
       }

       image = libraw_dcraw_make_mem_image(libraw_data, &rawErr);
       if (image == NULL)
       {
           libraw_close(libraw_data);
           return MakeGbmErrFromRawErr(rawErr, GBM_ERR_NOT_SUPP);
       }

       if (   (gbm->w   < image->width)
           || (gbm->h   < image->height)
           || (gbm->bpp != (image->colors * image->bits))
           || (image->type != LIBRAW_IMAGE_BITMAP)
           || ((gbm->bpp != 24) && (gbm->bpp != 48)) )
       {
           libraw_close(libraw_data);
           return GBM_ERR_READ;
       }

       /* copy data if requested */
       {
           const GBMRGB_16BPP backrgb = { 0, 0, 0 };
                 size_t stride_src = image->width * (gbm->bpp / 8);
           const size_t stride_dst = ((gbm->w * gbm->bpp + 31) / 32) * 4;
           const gbm_u8 * data8_src = (const gbm_u8 *) image->data;
                 gbm_u8 * data8_dst = data;
                 int      h;

           GBM * gbm_tmp = gbmmem_malloc(sizeof(GBM));
           if (gbm_tmp == NULL)
           {
              libraw_close(libraw_data);
              return GBM_ERR_MEM;
           }
           *gbm_tmp   = *gbm;
           gbm_tmp->w = image->width;
           gbm_tmp->h = image->height;

           if (isUnpackedHalfSize)
           {
               memset(data, 255, stride_dst * gbm->h);
           }

           /* flip vertically during reading */
           data8_src += stride_src * (gbm_tmp->h - 1);

           for (h = gbm_tmp->h - 1; h >= 0; --h)
           {
              if (! gbm_map_row_RGBx_BGRx(data8_src, gbm_tmp,
                                          data8_dst, gbm_tmp,
                                          &backrgb , GBM_FALSE))
              {
                  gbmmem_free(gbm_tmp);
                  libraw_close(libraw_data);
                  return GBM_ERR_NOT_SUPP;
              }
              data8_src -= stride_src;
              data8_dst += stride_dst;
           }
           gbmmem_free(gbm_tmp);
       }
       /* don't free the image here as this is internally done by libraw_close() */
   }

   libraw_close(libraw_data);
   return GBM_ERR_OK;
}

/* ---------------------------------------- */
/* ---------------------------------------- */

/* Read number of images in the PBM file. */
GBM_ERR raw_rimgcnt(const char *fn, int fd, int *pimgcnt)
{
   GBM_ERR rc;
   GBM gbm_src;
   memset(&gbm_src, 0, sizeof(GBM));

   fn = fn;  /* Suppress 'unref arg' compiler warnings */

   *pimgcnt = 0;
    rc = internal_raw_read(fd, &gbm_src, NULL, pimgcnt, GBM_TRUE);
    if ((rc == GBM_ERR_OK) && (*pimgcnt > 1))
    {
        *pimgcnt = 1;
    }
    return rc;
}

/* ---------------------------------------- */
/* ---------------------------------------- */

/* Execute rhdr() and rdata() in separate threads
 * with much more stack as libraw has huge stack
 * requirements and the GBM binary interface only
 * allows stack usage of up to 0x6000 bytes.
 */

#define LIBRAW_RESERVED_STACK (300*1024)

typedef struct
{
    int          fd;
    GBM        * gbm;
    const char * opt;
    GBM_ERR      rc;
} RHDR_THREAD_ARG;

static void Thread_raw_rhdr(GBM_THREAD_ARG arg)
{
   RHDR_THREAD_ARG * pArgs = (RHDR_THREAD_ARG *) arg;
   int imgcnt;
   RAW_PRIV_READ *raw_priv = (RAW_PRIV_READ *) pArgs->gbm->priv;

   /* copy possible options */
   if (strlen(pArgs->opt) >= sizeof(raw_priv->read_options))
   {
      pArgs->rc = GBM_ERR_BAD_OPTION;
      return;
   }
   strcpy(raw_priv->read_options, pArgs->opt);

   /* read bitmap info */
   pArgs->rc = internal_raw_read(pArgs->fd, pArgs->gbm, NULL, &imgcnt, GBM_FALSE);
}

GBM_ERR raw_rhdr(const char *fn, int fd, GBM *gbm, const char *opt)
{
   GBM_THREAD      threadId;
   RHDR_THREAD_ARG threadArgs;

   fn = fn;  /* Suppress 'unref arg' compiler warnings */

   threadArgs.fd  = fd;
   threadArgs.gbm = gbm;
   threadArgs.opt = opt;
   threadArgs.rc  = GBM_ERR_OK;

   threadId = gbmmthrd_scheduleTask(Thread_raw_rhdr, &threadArgs, LIBRAW_RESERVED_STACK);
   gbmmthrd_finishTask(&threadId);
   return threadArgs.rc;
}

/* ---------------------------------------- */
/* ---------------------------------------- */

GBM_ERR raw_rpal(int fd, GBM *gbm, GBMRGB *gbmrgb)
{
   fd=fd; gbm=gbm; gbmrgb=gbmrgb; /* Suppress 'unref arg' compiler warnings */
   return GBM_ERR_OK;
}

/* ---------------------------------------- */
/* ---------------------------------------- */

typedef struct
{
    int       fd;
    GBM     * gbm;
    gbm_u8  * data;
    GBM_ERR   rc;
} RDATA_THREAD_ARG;

static void Thread_raw_rdata(GBM_THREAD_ARG arg)
{
   RDATA_THREAD_ARG * pArgs = (RDATA_THREAD_ARG *) arg;
   int imgcnt;
   pArgs->rc = internal_raw_read(pArgs->fd, pArgs->gbm, pArgs->data, &imgcnt, GBM_FALSE);
}

GBM_ERR raw_rdata(int fd, GBM *gbm, gbm_u8 *data)
{
   GBM_THREAD       threadId;
   RDATA_THREAD_ARG threadArgs;

   threadArgs.fd   = fd;
   threadArgs.gbm  = gbm;
   threadArgs.data = data;
   threadArgs.rc   = GBM_ERR_OK;

   threadId = gbmmthrd_scheduleTask(Thread_raw_rdata, &threadArgs, LIBRAW_RESERVED_STACK);
   gbmmthrd_finishTask(&threadId);
   return threadArgs.rc;
}

/* ---------------------------------------- */
/* ---------------------------------------- */

const char *raw_err(GBM_ERR rc)
{
   switch ( (int) rc )
   {
      case GBM_ERR_RAW_NO_THUMB:
         return libraw_strerror(LIBRAW_NO_THUMBNAIL);

      case GBM_ERR_RAW_BAD_THUMB:
         return libraw_strerror(LIBRAW_UNSUPPORTED_THUMBNAIL);

      case GBM_ERR_RAW_BAD_QUALITY:
        return "quality is not in 0..3";

      case GBM_ERR_RAW_BAD_ABERRATION:
        return "aber= has unsupported values";

      case GBM_ERR_RAW_BAD_WB:
        return "wb= has unsupported values";

      case GBM_ERR_RAW_BAD_WBCOMB:
        return "avgwb and wb= cannot be used together";

      case GBM_ERR_RAW_BAD_BRIGHT:
        return "bright= has unsupported value";

      case GBM_ERR_RAW_BAD_BRIGHTCOMB:
        return "autobright and bright= cannot be used together";

      case GBM_ERR_RAW_BAD_DENOISE:
        return "denoise= has unsupported value";

      case GBM_ERR_RAW_BAD_MEDIANFILTER:
        return "medianfilter= has unsupported value (0..N)";
   }
   return NULL;
}

#endif // ENABLE_RAW

